
/************************************/
/*       Landscape Navigator        */
/*  by Daniele Raffo   31 JAN 2002  */
/*           DEA IFA UMLV           */
/* under GNU General Public License */
/************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <GL/glut.h>

#define FALSE 0
#define TRUE  1

/* landscape altitudes scale */
#define Z_SCALE 0.1


/* landscape image */
unsigned char *image = NULL;

/* dimensions of the landscape image */
int dx, dy;

/* landscape set of colors */
int antarctica = FALSE; 

/* land color */
GLfloat color[3];

/* texture data */
unsigned int *teximage;

/* mouse view coords */
GLdouble x00 = 0.0, y00 = 0.0;

/* some attributes of camera movement mode */
int jetpackon = FALSE;
int falling = FALSE;
int clipping = TRUE;

/* autospeed */
GLdouble speed = 0.0;

/* position of the camera */
GLdouble posx, posy, posz, alpha = 0.0;

/* window size */
int wxsize = 600, wysize = 400;

GLint id_landscape, id_plane, id_sky, id_moon;



/*******************/
/* Draws the scene */
/*******************/
static void drawFunc(void) { 
  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);

  glPushMatrix();

  /* camera position */
  gluLookAt(posx, posy, posz, 
	    posx + sin(alpha), posy + cos(alpha), posz,
  	    0.0, 0.0, 1.0); 

  /* shows the 3D landscape */
  glPushMatrix();
  glCallList(id_landscape);
  glPopMatrix(); 

  /* shows the plane where landscape will lay */
  glPushMatrix();
  glCallList(id_plane);
  glPopMatrix(); 

  /* shows the sky */
  glPushMatrix();
  glCallList(id_sky);
  glPopMatrix(); 

  /* shows the moon */
  glPushMatrix();
  glTranslatef(posx + dx * 1.95, posy + dy * 2.35, 255 * Z_SCALE * 2.33);
  glCallList(id_moon);
  glPopMatrix(); 
  
  glPopMatrix();

  glFlush();
  glutSwapBuffers();

}



/*****************/
/* Window resize */
/*****************/
static void reshapeFunc(int width,int height) { 

  GLfloat  h = (GLfloat) height / (GLfloat) width;
  glViewport(0, 0, (GLint) width, (GLint) height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-1.0, 1.0, -h, h, 1.0, 1200.0 );
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

}



/***************/
/* Mouse click */
/***************/
static void mouseFunc(int button, int state, int x, int y) { 
  if (button == GLUT_LEFT_BUTTON) {
    x00 = x; y00 = y;   /* new click: initializes new view coords at start */
  }
  
  glutPostRedisplay();

}



/************************/
/* Mouse click and hold */
/************************/
static void motionFunc(int x, int y) {

  alpha += 0.05 * (x - x00);
  posx -= 0.4 * sin(alpha) * (y - y00); 
  posy -= 0.4 * cos(alpha) * (y - y00);
  
  x00 = x; y00 = y;   /* difference of new actual view coords */

  force();
 
  glutPostRedisplay();
 
}



/************/
/* Keyboard */
/************/
static void kbdFunc(unsigned char c, int x, int y) {
  
  /* activates/deactivates jetpack */
  if (c == 'j') {
    switch (jetpackon) {
    case TRUE: jetpackon = FALSE; falling = TRUE; break;
    case FALSE: jetpackon = TRUE; falling = FALSE; break;
    }
  }
  
  /* sets autospeed */ 
  if (c >= 48 && c<= 57) speed = ((GLdouble) (c-48)) / 5; 
  
  /* change window size */
  if (c == '-') {
    if (wxsize > 300) {
      wxsize -= 300; wysize -= 200; 
      glutReshapeWindow(wxsize, wysize); 
      return; 
    }    
  }
  if (c == '+') {
    if (wxsize < 1800) {
      wxsize += 300; wysize += 200; 
      glutReshapeWindow(wxsize, wysize); 
      return; 
    }    
  }
  
  /* no clipping mode */
  if (c == 'X') clipping = FALSE;

  /* antarctica mode */
  if (c == 'A') {
    antarctica = !antarctica;
    init();
    glutPostRedisplay();
  }
  
  /* quit */
  if (c == 27) exit(0);  

}



/********************/
/* Keyboard special */
/********************/
static void kbdSpecialFunc(int code, int x, int y) { 
  
  if (code == GLUT_KEY_RIGHT) { alpha += 0.05; }
  if (code == GLUT_KEY_LEFT) { alpha -= 0.05; }
  if (code == GLUT_KEY_UP) { posx += 1.0 * sin(alpha); 
                             posy += 1.0 * cos(alpha); }
  if (code == GLUT_KEY_DOWN) { posx -= 1.0 * sin(alpha); 
                               posy -= 1.0 * cos(alpha); }

  if (jetpackon || !clipping) {
    if (code == GLUT_KEY_PAGE_UP) { posz += 0.7; }
    if (code == GLUT_KEY_PAGE_DOWN) { posz -= 0.7; }
  }
  
  force();

  glutPostRedisplay();

}



/********/
/* Idle */
/********/
static void idleFunc() {
  
  if (clipping && falling) {
    if (posz > imageval(integer(posx), integer(posy)) * Z_SCALE + 5) 
      posz -= 2;   /* falling to the ground */
    else
      falling = FALSE;   /* touched the ground */
  }
  
  if (speed > 0.0) {   /* no need to compute if speed is zero */   
    posx += speed * 1.0 * sin(alpha); 
    posy += speed * 1.0 * cos(alpha);
  }

  force();

  glutPostRedisplay();  

}



/******************************/
/* Reads a PPM landscape file */
/******************************/
unsigned char *readPPMland(char *filename) { 

  FILE *file;
  char header[255];
  unsigned char *img = NULL;
  unsigned char *dummy;
  int filetype; 
  enum {P5, P6, UNKNOWN};
  int c, idx = 0;
  
  if (!(file = fopen(filename, "rt"))) { 
    printf("Error: cannot find landscape file %s\n", filename);
    exit(1);
  }

  fscanf(file, "%s\n", header);
  if (!strcmp(header, "P5")) filetype = P5; 
  else if (!strcmp(header, "P6")) filetype = P6; 
  else filetype = UNKNOWN;
  do
    fgets(header, 255, file);
  while (header[0] == '#'); 
  sscanf(header,"%d %d",&dx,&dy);
  fscanf(file,"%s\n",header);
  img = (unsigned char *) malloc(dx * dy * sizeof(unsigned char));
  if (img == NULL) { 
    printf("Error allocating memory for PPM file\n");
    exit(1);
    }

  if (filetype == P5) {
    /* reads GRAY values */
    fread (img, sizeof (unsigned char), dx * dy, file);
  }
  else if (filetype == P6) {
    /* reads R G B values, throwing out the last two of three */
    for (c = 0; c < dx * dy; c++) {
      fread (&img[idx++], sizeof (unsigned char), 1, file);
      fread (dummy, sizeof (unsigned char), 2, file);
    }
  }
  else {
    printf("Error: unknown file type for %s\n", filename);
    exit(1);
  }

  fclose(file);
  
  return img;

}



/*********************************/
/* Forces camera to legal values */ 
/*********************************/
void force() {

  GLdouble ctop;   /* altitude of landscape below the camera, plus the gap */

  if (!clipping) return;

  /* forces camera inside the landscape map */
  if (posx < 0) posx = 0;
  if (posx > dx) posx = dx;
  if (posy < 0) posy = 0;
  if (posy > dy) posy = dy;
  
  ctop = imageval(integer(posx), integer(posy)) * Z_SCALE + 5;   /* 5 = gap */
  
  if ((!jetpackon) && (!falling)) posz = ctop;
  else if (jetpackon) if (posz < ctop) posz = ctop;
  
  if (posz > 255 * Z_SCALE * 2) posz = 255 * Z_SCALE * 2;

}



/************************************************************/
/* Returns the value at coords (x,y) on the landscape image */
/************************************************************/
unsigned char imageval(int x, int y) {

  /* two transformations are made: 
     1) finds a matrix coord from a monodimensional array 
     2) performs horiz image flipping by transforming x into dx-x-1 */
  return image[dx*y + dx-x-1];

}



/**************************************************/
/* Returns an interpolated value at coords (x, y) */
/**************************************************/
GLdouble iimageval(GLdouble ix, GLdouble iy) {

  /* warning: this function does not work. Explicit or implicit cast 
     of a GLdouble into a int or GLint will destroys graphics and result
     in a blank screen ! */
  
  GLdouble p1h = 0.0, p2h = 0.0, ph = 0.0;
  int refx = (int) ix, refy = (int) iy;
  
  p1h = (iy - (GLdouble) refy) * 
    (GLdouble) (imageval(refx, refy + 1) - imageval(refx, refy));
  p2h = (iy - (GLdouble) refy) * 
    (GLdouble) (imageval(refx + 1, refy + 1) - imageval(refx + 1, refy));
  if (p1h < 0) p1h = -p1h;   /* faster than a call to abs() */
  if (p2h < 0) p2h = -p2h;
  ph = (ix - refx) * (p2h - p1h); 
  if (ph < 0) ph = -ph;
  ph += (GLdouble) imageval(refx, refy);
  
  return ph;
  
}



/***************************************************************************/
/* Sets the color [0.0 - 1.0] of a point as func of its altitude [0 - 255] */
/***************************************************************************/
void setcolor(int altitude) {

  int lsb = altitude % 10;

  if (antarctica) {    
    /* this gives a nice antarctica landscape */
    GLfloat ptcolorR = (GLfloat) ((GLfloat) altitude * 1.00 / 255.0 + 0.00); 
    GLfloat ptcolorG = (GLfloat) ((GLfloat) altitude * 0.91 / 255.0 + 0.09); 
    GLfloat ptcolorB = (GLfloat) ((GLfloat) altitude * 0.64 / 255.0 + 0.36); 
    glColor3f(ptcolorR, ptcolorG, ptcolorB);
    return;
  }

  if (altitude > 0) {
    /* sand */
    color[0] = 0.98;
    color[1] = 0.95;
    color[2] = 0.70;
  }
  if (altitude > 2) {
    /* darker sand */
    color[0] = 0.95;
    color[1] = 0.90;
    color[2] = 0.65;
  }
  if (altitude > 10) {
    /* sand and green */
    color[0] = 0.75;
    color[1] = 0.85;
    color[2] = 0.41;
  }
  if (altitude > 20) {
    /* light green for vegetation */
    color[0] = 0.17;
    color[1] = 0.72 + lsb / 200.0;   /* adds some variation in color */ 
    color[2] = 0.16;
  }
  if (altitude > 90) {
    /* dark green for forests */
    color[0] = 0.13 + lsb / 200.0;
    color[1] = 0.58 - lsb / 100.0;
    color[2] = 0.15;
  }
  if (altitude > 155) {
    /* gray for stones */
    color[0] = 0.73 - lsb / 80.0;
    color[1] = 0.73 - lsb / 80.0;
    color[2] = 0.73 - lsb / 80.0;
  }
  if (altitude > 195) {
    /* white for ice */
    color[0] = 1.00;
    color[1] = 1.00;
    color[2] = 1.00;
  }

  glColor3fv(color);

}



/**********************************/
/* Returns the max of two numbers */
/**********************************/
int max(int a, int b) { return (a > b ? a : b); }



/*************************************/
/* Approximates a GLdouble to an int */
/*************************************/
int integer(GLdouble d) { return (int) (d + 0.5); }



/************************/
/* Scene initialization */ 
/************************/
static void init() {

  int i, j;
  int img_ij, img_i1j, img_ij1, img_i1j1;
  GLfloat txc = 2.0;   /* coords reference of texture source */
  GLfloat radius = max(dx, dy) * 3.0;   /* distance of the horizon */
  GLfloat skydim = radius * 1.0;   /* dimensions of the sky */
  GLfloat skyh = skydim * -0.2, skyht = skyh + skydim * 2;   /* sky position */
  GLUquadricObj *horizon, *moon;
  GLfloat fogcolor[] = {0.5, 0.5, 0.5, 1.0};
  
  glDisable(GL_LIGHTING);
  glClearColor(0.1, 0.1, 0.1, 0.0);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);


  /* builds the moon */
  id_moon = glGenLists(1);
  glNewList(id_moon, GL_COMPILE);
  glEnable(GL_TEXTURE_2D);  
  LoadTexture("landnav_tx_moon.rgb");
  glPushMatrix();
  moon = gluNewQuadric();
  gluQuadricTexture(moon, GL_TRUE);
  gluQuadricDrawStyle(moon, GLU_FILL);
  gluQuadricNormals(moon, GLU_FLAT);
  gluSphere(moon, 20, 200, 20);
  glPopMatrix();
  glDisable(GL_TEXTURE_2D);  
  glEndList();
  
  
  /* builds the sky */
  id_sky = glGenLists(1);
  glNewList(id_sky, GL_COMPILE);
  glPushMatrix();
  glEnable(GL_TEXTURE_2D);    
  LoadTexture("landnav_tx_skyn.rgb");
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); 
  glVertex3f(-skydim + dx/2, -skydim + dy/2, skyh);
  glTexCoord2f(0.0, txc); 
  glVertex3f(-skydim + dx/2, -skydim + dy/2, skyht);
  glTexCoord2f(txc, txc); 
  glVertex3f(-skydim + dx/2, skydim + dy/2, skyht);
  glTexCoord2f(txc, 0.0); 
  glVertex3f(-skydim + dx/2, skydim + dy/2, skyh);
  glEnd();
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); 
  glVertex3f(-skydim + dx/2, skydim + dy/2, skyh);
  glTexCoord2f(0.0, txc); 
  glVertex3f(-skydim + dx/2, skydim + dy/2, skyht);
  glTexCoord2f(txc, txc); 
  glVertex3f(skydim + dx/2, skydim + dy/2, skyht);
  glTexCoord2f(txc, 0.0); 
  glVertex3f(skydim + dx/2, skydim + dy/2, skyh);
  glEnd();
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); 
  glVertex3f(skydim + dx/2, skydim + dy/2, skyh);
  glTexCoord2f(0.0, txc); 
  glVertex3f(skydim + dx/2, skydim + dy/2, skyht);
  glTexCoord2f(txc, txc); 
  glVertex3f(skydim + dx/2, -skydim + dy/2, skyht);
  glTexCoord2f(txc, 0.0); 
  glVertex3f(skydim + dx/2, -skydim + dy/2, skyh);
  glEnd();
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); 
  glVertex3f(skydim + dx/2, -skydim + dy/2, skyh);
  glTexCoord2f(0.0, txc); 
  glVertex3f(skydim + dx/2, -skydim + dy/2, skyht);
  glTexCoord2f(txc, txc); 
  glVertex3f(-skydim + dx/2, -skydim + dy/2, skyht);
  glTexCoord2f(txc, 0.0); 
  glVertex3f(-skydim + dx/2, -skydim + dy/2, skyh);
  glEnd();
  glPopMatrix();
  glDisable(GL_TEXTURE_2D);  
  glEndList();


  /* builds the seascape (plane at altitude zero) */ 
  id_plane = glGenLists(1);
  glNewList(id_plane, GL_COMPILE);
  glEnable(GL_TEXTURE_2D);  
  LoadTexture("landnav_tx_sea.rgb");
  glPushMatrix();
  glTranslatef(dx/2, dy/2, 0.0);
  horizon = gluNewQuadric();
  gluQuadricTexture(horizon, GL_TRUE);
  gluQuadricDrawStyle(horizon, GLU_FILL);
  gluQuadricNormals(horizon, GLU_FLAT);
  gluDisk(horizon, 0.0, radius, 10, 10);  
  glPopMatrix();
  glDisable(GL_TEXTURE_2D);  
  glEndList();
  
  
  /* builds the 3D landscape */
  id_landscape = glGenLists(1);
  glNewList(id_landscape, GL_COMPILE);
  glPushMatrix();
  for (j = 0; j < dy - 1; j++) {
    for (i = 0; i < dx - 1; i++) {
      img_ij = imageval(i, j);
      img_i1j = imageval(i + 1, j);
      img_ij1 = imageval(i, j + 1);
      img_i1j1 = imageval(i + 1, j + 1);
      /* triangle 1 */
      /* if all three points are 0, don't draw the triangle 
	 to avoid covering the plane texture */
      if (img_ij + img_i1j + img_ij1 > 0) { 
	glBegin(GL_TRIANGLES);
	setcolor(img_ij);
	glVertex3d(i, j, img_ij * Z_SCALE);
	setcolor(img_i1j);
	glVertex3d(i+1, j, img_i1j * Z_SCALE);
	setcolor(img_ij1);
	glVertex3d(i, j+1, img_ij1 * Z_SCALE);
	glEnable(GL_SMOOTH);
	glEnd();
      }
      /* triangle 2 */
      /* if all three points are 0, don't draw the triangle 
	 to avoid covering the plane texture */
      if (img_i1j + img_ij1 + img_i1j1 > 0) { 
	glBegin(GL_TRIANGLES);
	setcolor(img_i1j);
	glVertex3d(i+1, j, img_i1j * Z_SCALE);
	setcolor(img_ij1);
	glVertex3d(i, j+1, img_ij1 * Z_SCALE);
	setcolor(img_i1j1);
	glVertex3d(i+1, j+1, img_i1j1 * Z_SCALE);
	glEnable(GL_SMOOTH);
	glEnd();
      }
    }
  }
  glPopMatrix();
  glEndList();
  
  glShadeModel(GL_SMOOTH);

}



/********************************************/
/* Fonctions for reading a RGB texture file */
/* Author: Roman Podobedov <romka@ut.ee>    */
/********************************************/    

/* texture file must be in .rgb (SGI) format, no compression, size 2^n */

void bwtorgba(unsigned char *b, unsigned char *l, int n) {

  while(n--) {
    l[0] = *b; l[1] = *b; l[2] = *b; l[3] = 0xff; 
    l += 4; b++;
  }

}



void latorgba(unsigned char *b, unsigned char *a, unsigned char *l, int n) {

  while(n--) {
    l[0] = *b; l[1] = *b; l[2] = *b; l[3] = *a; 
    l += 4; b++; a++;
  }

}



void rgbtorgba(unsigned char *r, unsigned char *g, unsigned char *b,
	       unsigned char *l, int n) {

  while(n--) {
    l[0] = r[0]; l[1] = g[0]; l[2] = b[0]; l[3] = 0xff; 
    l += 4; r++; g++; b++;
  }

}



void rgbatorgba(unsigned char *r, unsigned char *g, unsigned char *b,
		unsigned char *a, unsigned char *l, int n) {

  while(n--) {
    l[0] = r[0]; l[1] = g[0]; l[2] = b[0]; l[3] = a[0];
    l += 4; r++; g++; b++; a++;
  }
}



typedef struct _ImageRec {
  unsigned short imagic;
  unsigned short type;
  unsigned short dim;
  unsigned short xsize, ysize, zsize;
  unsigned int min, max;
  unsigned int wasteBytes;
  char name[80];
  unsigned long colorMap;
  FILE *file;
  unsigned char *tmp, *tmpR, *tmpG, *tmpB;
  unsigned long rleEnd;
  unsigned int *rowStart;
  int *rowSize;
} ImageRec;



static void ConvertShort(unsigned short *array, long length) {

  unsigned b1, b2;
  unsigned char *ptr;  
  ptr = (unsigned char *)array;
    while (length--) {
	b1 = *ptr++; b2 = *ptr++;
	*array++ = (b1 << 8) | (b2);
    }

}



static void ConvertLong(unsigned *array, long length) {

  unsigned b1, b2, b3, b4;
  unsigned char *ptr;
  ptr = (unsigned char *)array;

  while (length--) {
    b1 = *ptr++; b2 = *ptr++; b3 = *ptr++; b4 = *ptr++;
    *array++ = (b1 << 24) | (b2 << 16) | (b3 << 8) | (b4);
  }

}



static ImageRec *ImageOpen(const char *fileName) {

  union {
    int testWord;
    char testByte[4];
  } endianTest;

  ImageRec *image;
  int swapFlag;
  int x;
  endianTest.testWord = 1;

  if (endianTest.testByte[0] == 1) swapFlag = 1;
  else swapFlag = 0;
  image = (ImageRec *) malloc(sizeof(ImageRec));
  if (image == NULL) {
    fprintf(stderr, "Error: out of memory\n");
    exit(1);
  }

  if ((image->file = fopen(fileName, "rb")) == NULL) {
    perror(fileName);
    exit(1);
  }

  fread(image, 1, 12, image->file);
  if (swapFlag) ConvertShort(&image->imagic, 6);
  image->tmp = (unsigned char *) malloc(image->xsize*256);
  image->tmpR = (unsigned char *) malloc(image->xsize*256);
  image->tmpG = (unsigned char *) malloc(image->xsize*256);
  image->tmpB = (unsigned char *) malloc(image->xsize*256);

  if (image->tmp == NULL || image->tmpR == NULL || image->tmpG == NULL ||
      image->tmpB == NULL) {
    fprintf(stderr, "Error: out of memory\n");
    exit(1);
  }

  if ((image->type & 0xFF00) == 0x0100) {
    x = image->ysize * image->zsize * sizeof(unsigned);
    image->rowStart = (unsigned *) malloc(x);
    image->rowSize = (int *) malloc(x);
    if (image->rowStart == NULL || image->rowSize == NULL) {
      fprintf(stderr, "Error: out of memory\n");
      exit(1);
    }
    image->rleEnd = 512 + (2 * x);
    fseek(image->file, 512, SEEK_SET);
    fread(image->rowStart, 1, x, image->file);
    fread(image->rowSize, 1, x, image->file);
    if (swapFlag) {
      ConvertLong(image->rowStart, x/(int) sizeof(unsigned));
      ConvertLong((unsigned *) image->rowSize, x/(int) sizeof(int));
    }
  } else {
    image->rowStart = NULL;
    image->rowSize = NULL;
  }

  return image;

}



static void ImageClose(ImageRec *image) {

  fclose(image->file);
  free(image->tmp);
  free(image->tmpR);
  free(image->tmpG);
  free(image->tmpB);
  free(image->rowSize);
  free(image->rowStart);
  free(image);

}



static void ImageGetRow(ImageRec *image, unsigned char *buf, int y, int z) {

  unsigned char *iPtr, *oPtr, pixel;
  int count;  

  if ((image->type & 0xFF00) == 0x0100) {
    fseek(image->file, (long) image->rowStart[y+z*image->ysize], SEEK_SET);
    fread(image->tmp, 1, (unsigned int) image->rowSize[y+z*image->ysize],
	  image->file);
    iPtr = image->tmp;
    oPtr = buf;
    for (;;) {
      pixel = *iPtr++;
      count = (int)(pixel & 0x7F);
      if (!count) return; 
      if (pixel & 0x80) {
	while (count--) *oPtr++ = *iPtr++;
      } else {
	pixel = *iPtr++;
	while (count--) *oPtr++ = pixel;	
      }
    }
  } else {
    fseek(image->file, 512 + (y*image->xsize) + (z*image->xsize*image->ysize),
	  SEEK_SET);
    fread(buf, 1, image->xsize, image->file);
  }

}



unsigned *read_texture(char *name, int *width, int *height, int *components) {

  unsigned *base, *lptr;
  unsigned char *rbuf, *gbuf, *bbuf, *abuf;
  ImageRec *image;
  int y;

  image = ImageOpen(name);  
  if (!image) return NULL;
  (*width) = image->xsize;
  (*height) = image->ysize;
  (*components) = image->zsize;
  base = (unsigned *) malloc(image->xsize * image->ysize * sizeof(unsigned));
  rbuf = (unsigned char *) malloc(image->xsize * sizeof(unsigned char));
  gbuf = (unsigned char *) malloc(image->xsize * sizeof(unsigned char));
  bbuf = (unsigned char *) malloc(image->xsize * sizeof(unsigned char));
  abuf = (unsigned char *) malloc(image->xsize * sizeof(unsigned char));
  if(!base || !rbuf || !gbuf || !bbuf) return NULL;
  lptr = base;

  for (y=0; y < image->ysize; y++) {
    if (image->zsize >= 4) {
      ImageGetRow(image, rbuf, y, 0);
      ImageGetRow(image, gbuf, y, 1);
      ImageGetRow(image, bbuf, y, 2);
      ImageGetRow(image, abuf, y, 3);
      rgbatorgba(rbuf, gbuf, bbuf, abuf, (unsigned char *) lptr, image->xsize);
      lptr += image->xsize;
    } else if (image->zsize == 3) {
      ImageGetRow(image, rbuf, y, 0);
      ImageGetRow(image, gbuf, y, 1);
      ImageGetRow(image, bbuf, y, 2);
      rgbtorgba(rbuf, gbuf, bbuf, (unsigned char *) lptr, image->xsize);
      lptr += image->xsize;
    } else if (image->zsize == 2) {
      ImageGetRow(image, rbuf, y, 0);
      ImageGetRow(image, abuf, y, 1);
      latorgba(rbuf, abuf, (unsigned char *) lptr, image->xsize);
      lptr += image->xsize;
    } else {
      ImageGetRow(image, rbuf, y, 0);
      bwtorgba(rbuf, (unsigned char *) lptr, image->xsize);
      lptr += image->xsize;
    }
  }

  ImageClose(image);
  free(rbuf); free(gbuf); free(bbuf); free(abuf);  
  return (unsigned *) base;

}



void LoadTexture(char *fn) {

  int texwid, texht;
  int texcomps;

  teximage = read_texture(fn, &texwid, &texht, &texcomps);
  if (!teximage) {
    fprintf(stderr, "Error: cannot read texture file\n");
    exit(1);
  }
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, texwid, texht, 0, GL_RGBA, 
	       GL_UNSIGNED_BYTE, teximage);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glEnable(GL_TEXTURE_2D);  

} 



/********/
/* Main */
/********/
int main(int argc, char **argv) {

  /* do not put any printf() command here or OpenGl will core dump ! */

  if (argc != 2) {
    printf("Usage: landnav <ppm landscape file>\n");
    exit(1);
  }
  
  image = readPPMland(argv[1]);

  printf("-----------------------------------------------------\n");
  printf("  Landscape Navigator - an OpenGL landscape viewer\n");   
  printf("  by Daniele Raffo, 31 JAN 2002   DEA IFA UMLV\n");
  printf("  Freely use under the terms of the GNU GPL license\n");
  printf("-----------------------------------------------------\n\n");
  printf("Movement keys:\n");
  printf("    LEFT ARROW / RIGHT ARROW         turns left / right\n");
  printf("    UP ARROW / DOWN ARROW            moves forward / backward\n");
  printf("    You may also use the mouse by holding the left button and dragging\n\n");
  printf("    J                                jetpack on / off\n");
  printf("    PAGE UP / PAGE DOWN              climb / descend (when jetpack on)\n");
  printf("    0 ... 9                          automatic movement speed (0 stops)\n\n");
  printf("Funny effects:\n");
  printf("    A                                switches Antarctica mode\n");
  printf("    X                                no clipping\n\n");
  printf("Program:\n");
  printf("    + / -                            changes window size\n");
  printf("    ESC                              quit\n\n");

  /* positioning the camera at the center of landscape */
  posx = dx / 2, posy = dy / 2;

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowPosition(0, 0);
  glutInitWindowSize(wxsize, wysize);
  if (glutCreateWindow("Landscape Navigator") == GL_FALSE) return 1;
  
  init();  

  glutReshapeFunc(reshapeFunc);
  glutDisplayFunc(drawFunc);
  glutMouseFunc(mouseFunc);
  glutMotionFunc(motionFunc);
  glutKeyboardFunc(kbdFunc);
  glutSpecialFunc(kbdSpecialFunc);
  glutIdleFunc(idleFunc);

  glutMainLoop();
  
  return 0;

}


/* END OF FILE */







